"""
Email: autuanliu@163.com
Date: 2018/11/21

Ref:
1. Dynamic Granger causality based on Kalman filter for evaluation of functional network connectivity in fMRI data
2. https://stackoverflow.com/questions/3685265/how-to-write-a-multidimensional-array-to-a-text-file
"""

import datetime as dt

import numpy as np

__all__ = ['get_mat_data', 'normalize', 'save_2Darray', 'save_3Darray', 'Timer', 'make_linear_func', 'make_func4K4FROLS']


def get_mat_data(file_name, var_name):
    """从文件中读取出原始数据并转换为 np.array 类型

    Args:
        file_name (str): 数据存储的完整路径，如 'datastes/abc.mat'
        var_name (str): 存储数据的变量名

    Returns:
        np.array: 将读取到的原始数据转换为 np.ndarray 类型
    """

    import scipy.io as sio
    data_dict = sio.loadmat(file_name)
    return data_dict[var_name]


def normalize(data, scaler_type: str = 'MinMaxScaler'):
    """标准化数据

    Args:
        data (np.ndarray): 未经过标准化的原始数据
        scaler_type (str, optional): Defaults to 'MinMaxScaler'. 归一化方式
    """

    from inspect import isfunction
    from sklearn import preprocessing as skp
    if scaler_type in ['MinMaxScaler', 'StandardScaler']:
        data = getattr(skp, scaler_type)().fit_transform(data)
    elif isfunction(scaler_type):
        data = scaler_type(data)
    else:
        raise ValueError("""An invalid option was supplied, options are ['MinMaxScaler', 'StandardScaler', None] or lambda function.""")
    return data


def save_2Darray(file_path, data):
    """save np.array(2D) into txt file.(Ref2)

    Args:
        file_path (str or instance of Path(windowns or linux)): the file path to save data.
        data (np.array): the data need be saved.
    """

    with open(file_path, 'w') as outfile:
        outfile.write(f'# Array shape: {data.shape}\n')
        np.savetxt(outfile, data, fmt='%.4f')


def save_3Darray(file_path, data):
    """save np.array(3D) into txt file.(Ref2)

    Args:
        file_path (str or instance of Path(windowns or linux)): the file path to save data.
        data (np.array): the data need be saved.
    """

    with open(file_path, 'w') as outfile:
        # I'm writing a header here just for the sake of readability
        # Any line starting with "#" will be ignored by numpy.loadtxt
        outfile.write(f'# Array shape: {data.shape}\n')

        # Iterating through a ndimensional array produces slices along
        # the last axis. This is equivalent to data[i,:,:] in this case
        for data_slice in data:
            np.savetxt(outfile, data_slice, fmt='%.4f')
            # Writing out a break to indicate different slices...
            outfile.write('# New slice\n')


class Timer():
    """计时器类
    """

    def start(self):
        self.start_dt = dt.datetime.now()

    def stop(self):
        end_dt = dt.datetime.now()
        print(f'Time taken: {(end_dt - self.start_dt).total_seconds():.2f}s')


def make_linear_func(A_coef, var_name: str = 'y', step_name: str = 't', save=True, fname='est_model.txt', **kwargs):
    """根据系数生成模型表达式

    Args:
        A_coef (np.array): y = Ax + e 形式的模型系数矩阵 A
        var_name (str, optional): Defaults to 'y'. 使用的变量名
        step_name (str, optional): Defaults to 't'. 时间点变量名
        save (bool, optional): Defaults to True. 是否保存结果
        fname (str, optional): Defaults to 'est_model.txt'. 保存的文件名
        kwargs: np.savetxt args

    Returns:
        func_repr (np.array) 模型表达式
    """

    max_lag, n_dim, _ = A_coef.shape
    func_repr = []
    for var in range(n_dim):
        y = f'{var_name}{var+1}({step_name}) = '
        for dim in range(n_dim):
            for lag in range(max_lag):
                if abs(A_coef[lag, var, dim]) > 0.:
                    y += f'{A_coef[lag, var, dim]:.4f} * {var_name}{dim+1}({step_name}-{lag+1}) + '
        func_repr.append(y + f'e{var+1}({step_name})')
    func_repr = np.array(func_repr).reshape(-1, 1)
    if save:
        np.savetxt(fname, func_repr, fmt='%s', **kwargs)
    return func_repr


def make_func4K4FROLS(y_coef, terms_set, Kalman_S_No, var_name: str = 'x', step_name: str = 't', save=True, fname='est_model.txt', **kwargs):
    """生成使用 Kalman4FROLS 算法估计系数的模型表达式

    Args:
        y_coef (np.array): 估计的系数
        terms_set (np.array): 候选项集合
        Kalman_S_No (np.array): 候选项选择下标
        var_name (str, optional): Defaults to 'x'. 变量名
        step_name (str, optional): Defaults to 't'. 时间点变量名
        save (bool, optional): Defaults to True. 是否保存结果
        fname (str, optional): Defaults to 'est_model.txt'. 保存的文件名

    Returns:
        func_repr (np.array) 模型表达式
    """

    n_dim, n_term = y_coef.shape
    Kalman_S_No = np.sort(Kalman_S_No)
    func_repr = []
    for var in range(n_dim):
        y = f'{var_name}{var+1}({step_name}) = '
        for term in range(n_term):
            y += f'{y_coef[var, term]:.4f} * {terms_set[Kalman_S_No[var, term]]} + '
        func_repr.append(y + f'e{var+1}({step_name})')
    func_repr = np.array(func_repr).reshape(-1, 1)
    if save:
        np.savetxt(fname, func_repr, fmt='%s', **kwargs)
    return func_repr



"""基于 FROLS 算法的候选项选择器算法(通用算法，无关线性或者非线性)

**参考文献：
1. Billings S A, Chen S, Korenberg M J. Identification of MIMO non-linear systems using a forward-regression orthogonal estimator[J]. International Journal of Control, 1989, 49(6):2157-2189.
2. Billings S A. Nonlinear system identification : NARMAX methods in the time, frequency, and spatio-temporal domains[M]. Wiley, 2013.

**Notes: 参看 matlab 相关实现以及其保存数据

Copyright:
----------
    Author: AutuanLiu
    Date: 2018/12/10
"""

import numpy as np

from .utils import get_mat_data

__all__ = ['Selector']


class Selector:
    """基于 FROLS 算法的候选项选择器算法(通用算法，无关线性或者非线性)

    Attributes:
        norder (int): 非线性次数
        max_lag (int): max lag.
        n_dim (int): the channel or dim of signals.
        Kalman_H (np.array): 供 kalman 滤波器使用的候选项矩阵, 参看 matlab 代码实现
        normalized_signals (np.array): 标准化之后的信号数据, 参看 matlab 代码实现
        Hv (np.array): 基于 base(linear terms) 的候选项组合, 参看 matlab 代码实现
        S_No (np.array): sparse matrix, 表示选择候选项的下标或索引, 参看 matlab 代码实现
        candidate_terms (np.array): 候选项集合
        Kalman_S_No (np.array): 和 Kalman_H 相匹配的候选项选择下标
        n_term (int): FROLS 的阈值或者需要选择的候选项个数
    """

    def __init__(self, terms_path):
        """基于 FROLS 算法的候选项选择器算法

        Args:
            terms_path (str): term selector(matlab) 程序的结果路径
        """

        for key in ['normalized_signals', 'Hv', 'Kalman_H', 'S_No']:
            setattr(self, key, get_mat_data(terms_path, key))
        self.n_dim, n_point, self.n_term = self.Kalman_H.shape
        self.norder = self.Hv.shape[0]
        self.max_lag = self.normalized_signals.shape[0] - n_point
        self.candidate_terms = None
        # !这里注意消除 matlab 和 python 之间的索引差异
        self.Kalman_S_No = self.S_No[:, :self.n_term] - 1

    def make_terms(self, var_name: str = 'x', step_name: str = 't'):
        """生成模型候选项表达式

        Args:
            var_name (str, optional): Defaults to 'x'. 使用的变量名
            step_name (str, optional): Defaults to 't'. 时间点变量名

        Returns:
            terms_repr (np.array) 模型表达式
        """

        terms_repr = []
        base = []
        nonlinear = []
        for var in range(self.n_dim):
            for lag in range(self.max_lag):
                base.append(f'{var_name}{var+1}({step_name}-{lag+1})')
        base = np.asarray(base)
        if self.norder == 1:
            terms_repr = base
        else:
            for ord in range(2, self.norder + 1):
                Hv_tmp = self.Hv[ord - 1, 0]
                n_row, n_var = Hv_tmp.shape
                for m in range(n_row):
                    term = ''
                    for n in range(n_var):
                        # !这里注意消除 matlab 和 python 之间的索引差异
                        term += f'{base[Hv_tmp[m, n]-1]}*'    # *matlab 索引从1开始
                    nonlinear.append(term[:-1])
            nonlinear = np.asarray(nonlinear)
            tmp = list(base)
            tmp.extend(list(nonlinear))
            terms_repr = np.asarray(tmp)
        self.candidate_terms = terms_repr    # 更新 candidate_terms
        return terms_repr

    def make_selection(self):
        """生成与模型候选项选择有关的数据执行此方法前，先执行 make_terms 方法

        Args:
            var_name (str, optional): Defaults to 'x'. 使用的变量名
            step_name (str, optional): Defaults to 't'. 时间点变量名

        Returns:
            normalized_signals (np.array): 标准化之后的信号数据
            Kalman_H (np.array): 供 kalman 滤波器使用的候选项矩阵
            candidate_terms (np.array): 候选项集合
            Kalman_S_No (np.array): 和 Kalman_H 相匹配的候选项选择下标
        """

        return self.normalized_signals, self.Kalman_H, self.candidate_terms, self.Kalman_S_No

    def __repr__(self):
        return f'Selector(norder={self.norder}, max_lag={self.max_lag}, n_dim={self.n_dim}, n_term={self.n_term})'



"""
Email: autuanliu@163.com
Date: 2018/12/12
"""

from .estimator import *
from .regression import *
from .selector import *
from .tools import *
from .utils import *

__all__ = [
    'Kalman4ARX', 'Kalman4FROLS', 'torch4FROLS', 'regression4torch', 'regression4sklearn', 'TermsData', 'Selector', 'grid_search1', 'grid_search2',
    'grid_search3', 'plot_grid_search', 'get_mat_data', 'normalize', 'save_2Darray', 'save_3Darray', 'Timer', 'make_linear_func', 'make_func4K4FROLS',
    'make_dataset4SK', 'Linear'
]



"""
Email: autuanliu@163.com
Date: 2018/11/212
# !!用于网格搜索
"""

import numpy as np

from .estimator import Kalman4ARX

__all__ = ['grid_search1', 'grid_search2', 'grid_search3', 'plot_grid_search']


def grid_search1(signals, lag_range, uc_range):
    """对超参数进行网格搜索策略(policy 1)，lag 和 uc 均不固定

    Args:
        signals (np.array): the data need to process.
        lag_range (list or tuple or np.array, iterable): the range of max_lag.
        uc_range (list or tuple or np.array, iterable): the range of update coefficient.

    Returns:
        best_lag_uc
    """

    # policy 1: (max_lag, uc) 的最佳组合，使得 mse_loss 最小
    min_mse = 100
    best_lag_uc = (5, 0.001)
    for m in lag_range:
        for n in uc_range:
            kf = Kalman4ARX(signals, m, n)
            _ = kf.smoother()
            min_mse = kf.mse_loss if kf.mse_loss <= min_mse else min_mse
            print(f'max_lag={m}, uc={n}, mse_loss={kf.mse_loss}')
            best_lag_uc = (m, n)
    return best_lag_uc


def grid_search2(signals, lag_range, uc, criterion='BIC', plot=False):
    """对超参数进行网格搜索策略(policy 2)，uc 固定

    Args:
        signals (np.array): the data need to process.
        lag_range (list or tuple or np.array, iterable): the range of max_lag.
        uc (float): update coefficient.
        criterion (str): the criterion for searching max_lag.
        plot (bool): 是否绘图

    Returns:
        best_lag, (x, y)
    """

    # policy 2: max_lag的最佳值，使得 AIC 或者 BIC 最小
    min_criterion = 100
    best_lag = 1
    # 绘图时保存
    x, y = [], []
    for m in lag_range:
        kf = Kalman4ARX(signals, m, uc)
        _ = kf.smoother()
        min_criterion = getattr(kf, criterion.lower()) if getattr(kf, criterion.lower()) <= min_criterion else min_criterion
        print(f'max_lag={m}, {criterion}={getattr(kf, criterion.lower())}')
        best_lag = m
        x.append(m)
        y.append(getattr(kf, criterion.lower()))
    if plot:
        plot_grid_search(x, y, 'max_lag', criterion, f'max_lag2{criterion}')
    return best_lag, (x, y)


def grid_search3(signals, max_lag, uc_range, criterion='BIC', plot=False):
    """对超参数进行网格搜索策略(policy 3)，max_lag 固定

    Args:
        signals (np.array): the data need to process.
        max_lag (int): max_lag.
        uc_range (list or tuple or np.array, iterable): the range of update coefficient.
        criterion (str): the criterion for searching max_lag.
        plot (bool): 是否绘图

    Returns:
        best_uc, (x, y)
    """

    # policy 2: uc 的最佳值，使得 mse_loss 最小
    min_mse = 100
    best_uc = 1
    # 绘图时保存
    x, y = [], []
    for m in uc_range:
        kf = Kalman4ARX(signals, max_lag, m)
        _ = kf.smoother()
        min_mse = kf.mse_loss if kf.mse_loss <= min_mse else min_mse
        print(f'uc={m}, mse_loss={kf.mse_loss}')
        best_uc = m
        x.append(m)
        y.append(kf.mse_loss)
    if plot:
        plot_grid_search(x, y, 'uc', 'mse_loss', 'uc2mse_loss')
    return best_uc, (x, y)


def plot_grid_search(x, y, x_label, y_label, file_name):
    import matplotlib.pyplot as plt
    plt.plot(x, y, 'b+')
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    # save figure
    plt.savefig(f'./kalman_filter/images/{file_name}.png')



"""继承 filterpy.kalman 模块中的 KalmanFilter，并覆盖 update 方法，以满足估计系数的需要，因为在估计系数时需要同时更新 Q 与 R。

1. https://filterpy.readthedocs.io/en/latest/kalman/KalmanFilter.html
文献：
1. Accurate epileptogenic focus localization through time-variant functional connectivity analysis of intracranial electroencephalographic signals
2. A new Kalman filter approach for the estimation of high-dimensional time-variant multivariate AR models and its application in analysis of laser-evoked brain potentials
3. Dynamic Granger causality based on Kalman filter for evaluation of functional network connectivity in fMRI data
4. Seizure-Onset Mapping Based on Time-Variant Multivariate Functional Connectivity Analysis of High-Dimensional Intracranial EEG: A Kalman Filter Approach
5. Nonlinear system identification, NARMAX methods in the time, frequency, and spation-temporal domains, Stephen A Billings, 2013,
6. Identification of MIMO non-linear systems using a forward -regression orthogonal estimator, International Journal of Control, Billings, S A, Chen, S, Korenberg, M J, 1989

Copyright:
----------
    Author: AutuanLiu
    Date: 2018/11/21(linear), 2018/12/6(nonlinear)
"""

from copy import deepcopy

import numpy as np
from filterpy.common import reshape_z
from filterpy.kalman import KalmanFilter
from numpy import dot, eye, zeros

from .regression import (TermsData, make_dataset4SK, regression4sklearn, regression4torch)

__all__ = ['Kalman4ARX', 'Kalman4FROLS', 'torch4FROLS']


class Kalman4ARX(KalmanFilter):
    """定义适用于估计ARX时间序列模型系数的 kalman filter。

    Attributes:
        max_lag (int): max lag of model.
        signals (np.array): N * ndim. normalized_signals 经过标准化的数据
        N (int): 信号的长度
        ndim (int): 信号的维数
        uc (float): update coefficient.
        aic (float): Akaike information criterion.
        bic (float): Bayesian information criterion.
        mse_loss (float): mean square error loss.
    """

    def __init__(self, signals, max_lag=3, uc=0.0001):
        """构造函数。

        Args:
            signals (np.array): 可观测信号(n_point*n_dim)
            max_lag (int, optional): Defaults to 3. 自回归模型的阶数，即最大延迟
            uc (float, optional): Defaults to 0.0001. update coefficient, forgetting factor.
        """

        n_row, n_col = signals.shape
        dim_x = max_lag * (n_col**2)
        super().__init__(dim_x, n_col, dim_u=0)
        self.max_lag = max_lag
        self.signals = signals
        self.N = n_row    # 信号的长度
        self.ndim = n_col    # 信号的维数
        self.uc = uc    # update coefficient
        self.z_s = 0    # 滤波器最后得到的观测值估计序列
        self.init()

    def init(self):
        self.x = np.random.randn(self.dim_x, 1)    # 初始状态初始化为 (0, 1) 正态分布
        self.Q = self.uc * eye(self.dim_x)    # 文献1的初始化方式，若使用文献3的初始化方式，注释掉该行
        self.H = self.measurement_matrix(self.max_lag)    # 初始时的测量矩阵
        self.z = self.signals[self.max_lag].reshape(-1, 1)    # 初始时的测量值

    def measurement_matrix(self, time):
        """计算 C_n, 参考文献3

        Args:
            time: int, 当前的时间点, 从 0 开始

        Returns:
            measurement_matrix: np.array, 与当前时刻对应的转移矩阵
        """

        Yt = lambda t: self.signals[(t - 1)::-1, :] if t == self.max_lag else self.signals[(t - 1):(t - 1 - self.max_lag):-1, :]
        Y = Yt(time).reshape(1, -1)
        Cn = np.kron(eye(self.ndim), Y.T)    # 这里其实计算的是 n-1 时刻的 C
        return Cn.T

    def update(self, z, R=None, H=None):
        """与原update函数的唯一不同在于P更新的方式可以调节，为了保证其它功能正常，仍然保留原始的内容
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = np.zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        if R is None:
            R = self.R
        elif np.isscalar(R):
            R = eye(self.dim_z) * R

        if H is None:
            H = self.H

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - dot(H, self.x)

        # common subexpression for speed
        PHT = dot(self.P, H.T)

        # S = HPH' + R
        # project system uncertainty into measurement space
        self.S = dot(H, PHT) + R
        self.SI = self.inv(self.S)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = dot(PHT, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + dot(self.K, self.y)

        # P = (I-KH)P(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P = (I-KH)P usually seen in the literature.

        I_KH = self._I - dot(self.K, H)
        # self.P = dot(I_KH, self.P)   # 通常情况下的处理
        self.P = dot(dot(I_KH, self.P), I_KH.T) + dot(dot(self.K, R), self.K.T)

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def filter(self, time, z):
        """实现一次滤波
        使用 self.x 获取当前的预测值

        Args:
            time (int): 当前时间点
            z (np.array): column vector, 当前观测值

        Returns:
            z_s (np.array): column vector, 当前的预测值
        """

        self.predict()
        self.update_Q()
        self.update_R(z)
        self.update_H(time)
        self.update(z.T)
        return self.z

    def forward(self):
        """滤波器的前向操作。

        Returns:
            x: (np.array) 经过滤波器后状态的最终估计值
            P: (np.array) 经过滤波器后最终的预测误差
            z_s: (np.array) 经过滤波器后观测值的估计值
        """

        z_s = []
        for time, z in enumerate(self.signals[self.max_lag:]):
            z_s.append(self.filter(time + self.max_lag, z.reshape(-1, 1)))
        self.z_s = np.array(z_s).squeeze()
        return self.x, self.P, self.z_s

    def backward(self):
        """滤波器的后向操作。这里使用同一个滤波器先进行前行操作，之后进行后向操作的连续过程，参看
        smoother 方法，避免使用两个滤波器，在进行后向操作还要使用前向操作的最终状态进行初始化的问题。

        Returns:
            x: (np.array) 经过滤波器后状态的最终估计值
            P: (np.array) 经过滤波器后最终的预测误差
            z_s: (np.array) 经过滤波器后观测值的估计值
        """

        z_s = []
        for time, z in enumerate(self.signals[:(self.max_lag - 1):-1]):
            z_s.append(self.filter(self.N - 1 - time, z.reshape(-1, 1)))
        self.z_s = np.array(z_s).squeeze()
        return self.x, self.P, self.z_s[::-1]

    def smoother(self):
        """根据 forward 和 backward 得到的数值进行光滑处理，参考文献 3
        """

        x_f, P_f, _ = self.forward()
        x_b, P_b, _ = self.backward()
        factor1 = self.inv(self.inv(P_f) + self.inv(P_b))
        factor2 = dot(self.inv(P_f), x_f) + dot(self.inv(P_b), x_b)
        x_s = dot(factor1, factor2)
        return x_s

    def update_R(self, z):
        # update R, 文献 3
        residual = self.residual_of(z)
        self.R = (1 - self.uc) * self.R + self.uc * dot(residual, residual.T)

    def update_Q(self):
        # update Q, 文献 3
        self.Q = self.uc * self._I * np.trace(self.P_prior) / (self.max_lag * self.N)

        # update Q, 文献 1
        # self.Q = self.uc * self._I

    def update_H(self, time):
        # update H, 文献1, 3, 测量矩阵会随时间变化
        self.H = self.measurement_matrix(time)

    def estimate_coef(self, threshold=1e-5):
        """重新排列计算出来的系数(pK^2 x 1)

        Args:
            threshold: 将系数置 0 的阈值, 默认为 1e-5

        Returns:
            y_coef: 以信号形式排列的系数
            A_coef: 以系数矩阵形式排列的系数
        """

        A_coef = []
        x_s = self.smoother()
        x_s[np.abs(x_s) < threshold] = 0.    # 阈值处理
        y_coef = x_s.T.reshape(self.ndim, -1)    # 重新排列系数为 ndim x (ndim * p)

        # A_coef 为 p x (ndim x ndim) 形式矩阵
        for m in range(0, y_coef.shape[1], self.ndim):
            A_coef += [y_coef[:, m:(m + self.ndim)]]
        return y_coef, np.array(A_coef)

    def AIC(self, p):
        """Akaike information criterion.
        A major concern with parametric analysis methods is the order selection of the autoregressive (AR) model.
        If the order is too small, the frequency content in the data cannot be resolved and the spectral estimates will be biased and smoothed and consequently,
        some information will be lost. On the other hand, if the model order is overly large, spurious peaks (instabilities) can occur in the spectral estimates,
        which result in a large variance of error and might be misleading thus causing wrong identification. (reference 3)

        AIC(p) = ln(det(R)) + 2p*d^2 / T

        Returns:
            the AIC value of self.max_lag
        """

        aic = np.log(np.linalg.det(self.R)) + 2 * p * (self.ndim**2) / self.N
        return aic

    def BIC(self, p):
        """Bayesian information criterion.
        p is the model order, T is the length of time series, R is the covariance matrix of the measurement noise,
        and d is the number of the time series under investigation. Both of these criteria using maximum likelihood principle make a compromise
        between model complexity and goodness of fit and track both the decreasing error power and the increasing spectral variance with respect to
        an increasing model order. (reference 3)

        BIC(p) = ln(det(R)) + ln(T)*p*d^2 / T

        Returns:
            the BIC value of self.max_lag
        """

        bic = np.log(np.linalg.det(self.R)) + np.log(self.N) * p * (self.ndim**2) / self.N
        return bic

    @property
    def aic(self):
        return self.AIC(self.max_lag)

    @property
    def bic(self):
        return self.BIC(self.max_lag)

    @property
    def mse_loss(self):
        return np.sum((self.z_s - self.signals[self.max_lag:])**2) / (self.N - self.max_lag)


class Kalman4FROLS(KalmanFilter):
    """定义适用于估计NARX时间序列模型系数的 kalman filter。

    Attributes:
        max_lag (int): max lag of model.
        signals (np.array): 输入的信号数据 normalized_signals 经过标准化的数据
        Kalman_H (np.array): 测量矩阵 measurement matrix
        N (int): 信号的有效长度
        ndim (int): 信号的维数
        uc (float): update coefficient.
    """

    def __init__(self, signals, Kalman_H, uc=0.0001):
        """构造函数。

        Args:
            signals (np.array): 可观测信号(n_point*n_dim) (N 的大小和 Kalman_H 的行数相同)
            Kalman_H (np.array): 测量矩阵 measurement matrix
            uc (float, optional): Defaults to 0.0001. update coefficient, forgetting factor.
        """

        max_lag = signals.shape[0] - Kalman_H.shape[1]
        n_point, n_dim = Kalman_H.shape[1], Kalman_H.shape[0]
        dim_x = n_dim * Kalman_H.shape[2]
        super().__init__(dim_x, n_dim, dim_u=0)
        self.Kalman_H = Kalman_H
        self.max_lag = max_lag
        self.signals = signals
        self.N = n_point    # 信号的有效长度
        self.ndim = n_dim    # 信号的维数
        self.uc = uc    # update coefficient
        self.z_s = None    # 滤波器最后得到的观测值估计序列
        self.init()

    def init(self):
        self.x = np.random.randn(self.dim_x, 1)    # 初始状态初始化为 (0, 1) 正态分布
        self.Q = self.uc * eye(self.dim_x)    # 文献1的初始化方式，若使用文献3的初始化方式，注释掉该行
        self.H = self.measurement_matrix(0)    # 初始时的测量矩阵
        self.z = self.signals[self.max_lag].reshape(-1, 1)    # 初始时的测量值

    def measurement_matrix(self, time):
        """计算 C_n, 参考文献3

        Args:
            time: int, 当前的时间点, 从 0 开始

        Returns:
            measurement_matrix: np.array, 与当前时刻对应的转移矩阵
        """

        Cn = np.kron(eye(self.ndim), self.Kalman_H[:, time, :])    #! 这里其实计算的是 n 时刻的 C
        idx = slice(0, Cn.shape[0], self.ndim + 1)
        return Cn[idx]

    def update(self, z, R=None, H=None):
        """与原update函数的唯一不同在于P更新的方式可以调节，为了保证其它功能正常，仍然保留原始的内容"""

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = np.zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        if R is None:
            R = self.R
        elif np.isscalar(R):
            R = eye(self.dim_z) * R

        if H is None:
            H = self.H

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - dot(H, self.x)

        # common subexpression for speed
        PHT = dot(self.P, H.T)

        # S = HPH' + R
        # project system uncertainty into measurement space
        self.S = dot(H, PHT) + R
        self.SI = self.inv(self.S)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = dot(PHT, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + dot(self.K, self.y)

        # P = (I-KH)P(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P = (I-KH)P usually seen in the literature.

        I_KH = self._I - dot(self.K, H)
        # self.P = dot(I_KH, self.P)   # 通常情况下的处理
        self.P = dot(dot(I_KH, self.P), I_KH.T) + dot(dot(self.K, R), self.K.T)

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def filter(self, time, z):
        """实现一次滤波
        使用 self.x 获取当前的预测值

        Args:
            time (int): 当前时间点
            z (np.array): column vector, 当前观测值

        Returns:
            z_s (np.array): column vector, 当前的预测值
        """

        self.predict()
        self.update_Q()
        self.update_R(z)
        self.update_H(time)
        self.update(z.T)
        return self.z

    def forward(self):
        """滤波器的前向操作。

        Returns:
            x: (np.array) 经过滤波器后状态的最终估计值
            P: (np.array) 经过滤波器后最终的预测误差
            z_s: (np.array) 经过滤波器后观测值的估计值
        """

        z_s = []
        for time, z in enumerate(self.signals[self.max_lag:]):
            z_s.append(self.filter(time, z.reshape(-1, 1)))
        self.z_s = np.array(z_s).squeeze()
        return self.x, self.P, self.z_s

    def backward(self):
        """滤波器的后向操作。这里使用同一个滤波器先进行前行操作，之后进行后向操作的连续过程，参看
        smoother 方法，避免使用两个滤波器，在进行后向操作还要使用前向操作的最终状态进行初始化的问题。

        Returns:
            x: (np.array) 经过滤波器后状态的最终估计值
            P: (np.array) 经过滤波器后最终的预测误差
            z_s: (np.array) 经过滤波器后观测值的估计值
        """

        z_s = []
        for time, z in enumerate(self.signals[:(self.max_lag - 1):-1]):
            z_s.append(self.filter(self.N - 1 - time, z.reshape(-1, 1)))
        self.z_s = np.array(z_s).squeeze()
        return self.x, self.P, self.z_s[::-1]

    def smoother(self):
        """根据 forward 和 backward 得到的数值进行光滑处理，参考文献 3
        """

        x_f, P_f, _ = self.forward()
        x_b, P_b, _ = self.backward()
        factor1 = self.inv(self.inv(P_f) + self.inv(P_b))
        factor2 = dot(self.inv(P_f), x_f) + dot(self.inv(P_b), x_b)
        x_s = dot(factor1, factor2)
        return x_s

    def update_R(self, z):
        # update R, 文献 3
        residual = self.residual_of(z)
        self.R = (1 - self.uc) * self.R + self.uc * dot(residual, residual.T)

    def update_Q(self):
        # update Q, 文献 3
        self.Q = self.uc * self._I * np.trace(self.P_prior) / (self.max_lag * self.N)

        # update Q, 文献 1
        # self.Q = self.uc * self._I

    def update_H(self, time):
        # update H, 文献1, 3, 测量矩阵会随时间变化
        self.H = self.measurement_matrix(time)

    def estimate_coef(self, threshold=1e-5):
        """重新排列计算出来的系数(pK^2 x 1)

        Args:
            threshold: 将系数置 0 的阈值, 默认为 1e-5

        Returns:
            y_coef: 以信号形式排列的系数
            A_coef: 以系数矩阵形式排列的系数
        """

        x_s = self.smoother()
        x_s[np.abs(x_s) < threshold] = 0.    # 阈值处理
        y_coef = x_s.T.reshape(self.ndim, -1)    # 重新排列系数为 ndim x (ndim * p)
        return y_coef


class torch4FROLS:
    def __init__(self, signals, Kalman_H, n_epoch=50, batchsize=32, learning_rate=0.001):
        """构造函数

        Args:
            signals (np.array): normalized_signals 经过标准化的数据
            Kalman_H (np.array): Kalman 候选项矩阵
            n_epoch(int, optional): Defaults to 32. 训练 epoch 数
            batchsize(int, optional): Defaults to 32.
            learning_rate(int, optional): Defaults to 0.001.
        """

        import torch
        from .regression import regression4torch, TermsData
        from torch.utils.data import DataLoader
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        self.n_dim = Kalman_H.shape[0]
        data = TermsData(signals, Kalman_H)
        in_dim = Kalman_H.shape[0] * Kalman_H.shape[2]
        out_dim = Kalman_H.shape[0]
        self.n_epoch = n_epoch
        self.loader = DataLoader(dataset=data, batch_size=batchsize, shuffle=True)
        self.model = regression4torch(in_dim, out_dim).to(self.device)
        self.criterion = torch.nn.MSELoss()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate)

    def estimate_coef(self):
        """估计模型系数
        """

        self.model.train()

        def _train():
            for batch_idx, (data, target) in enumerate(self.loader):
                data, target = data.to(self.device), target.to(self.device)
                self.optimizer.zero_grad()
                output = self.model(data)
                loss = self.criterion(output, target)
                loss.backward()
                self.optimizer.step()
            print(f'loss: {loss}')

        for _ in range(self.n_epoch):
            _train()
        return self.model.weight.detach().cpu().numpy().reshape(self.n_dim, -1)


# !原生代码和此处的问题有区别，暂时不可直接使用
# class sklearn4FROLS:
#     """基于 scikit-learn 的模型系数估计(原生代码和此处的问题有区别，暂时不可直接使用)
#     """

#     def __init__(self, signals, Kalman_H):
#         """构造函数

#         Args:
#             signals (np.array): normalized_signals 经过标准化的数据
#             Kalman_H (np.array): Kalman 候选项矩阵
#         """

#         from .regression import regression4sklearn, make_dataset4SK
#         max_lag = signals.shape[0] - Kalman_H.shape[1]
#         self.data, self.target = make_dataset4SK(signals, Kalman_H)
#         self.model = regression4sklearn()

#     def estimate_coef(self):
#         """估计模型系数
#         """

#         # 训练模型
#         self.model.fit(self.data, self.target)
#         # 获取模型参数
#         return self.model.coef_.reshape(self.target.shape[1], -1)



"""setup

Copyright:
----------
    Author: AutuanLiu
    Date: 2019/01/08
"""

import distutils.spawn
import shlex
import subprocess
import sys

from setuptools import find_packages, setup

version = '0.3.2'

if sys.argv[1] == 'release':
    if not distutils.spawn.find_executable('twine'):
        print(
            'Please install twine:\n\n\tpip install twine\n',
            file=sys.stderr,
        )
        sys.exit(1)

    commands = [
        'git pull origin master',
        'git tag v{:s}'.format(version),
        'git push origin master --tag',
        'python setup.py sdist',
        'twine upload dist/imgviz-{:s}.tar.gz'.format(version),
    ]
    for cmd in commands:
        print('+ {}'.format(cmd))
        subprocess.check_call(shlex.split(cmd))
    sys.exit(0)


def get_install_requires():
    install_requires = []
    with open('requirements.txt') as f:
        for req in f:
            install_requires.append(req.strip())
    return install_requires


with open('description.md') as f:
    long_description = f.read()

setup(
    name='kalman-estimation',
    version=version,
    packages=find_packages(),
    install_requires=get_install_requires(),
    description='Kalman filter based coefficient estimation toolbox.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    python_requires='>=3.5',
    author='Autuan Liu',
    author_email='autuanliu@163.com',
    url='https://github.com/AutuanLiu/Kalman-Filter',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



"""
Email: autuanliu@163.com
Date: 2018/12/12
"""

import math

import numpy as np
import torch
from torch.nn import Module, Parameter, init
from torch.utils.data import Dataset

__all__ = ['Linear', 'regression4torch', 'regression4sklearn', 'TermsData', 'make_dataset4SK']


class Linear(Module):
    r"""Applies a linear transformation to the incoming data: :math:`y = x@A + b`

    REF: https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/linear.py#L11

    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        bias: If set to False, the layer will not learn an additive bias.
            Default: ``True``
            
    Attributes:
        weight: the learnable weights of the module
        bias:   the learnable bias of the module
    """

    __constants__ = ['bias']

    def __init__(self, in_features, out_features, bias=True):
        super(Linear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.Tensor(in_features))
        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        bound = 1 / math.sqrt(self.in_features)
        init.uniform_(self.weight, -bound, bound)
        if self.bias is not None:
            init.uniform_(self.bias, -bound, bound)

    def forward(self, input):
        if self.bias is not None:
            return input @ self.weight + self.bias
        else:
            return input @ self.weight

    def extra_repr(self):
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}'


def regression4torch(in_dim, out_dim):
    """基于 PyTorch 的线性回归模型 Y = X*W +b
    
    原有的 nn.Linear() 不适用于这里的问题，所以定义 Linear 如上

    Args:
        in_dim (int): 输入维度
        out_dim (int): 输出维度
    """

    return Linear(in_dim, out_dim, bias=False)


def regression4sklearn():
    """基于 scikit-learn 的线性回归模型
    """

    from sklearn.linear_model import LinearRegression
    return LinearRegression(fit_intercept=False)


class TermsData(Dataset):
    """构造关于候选项的数据集
    
    Attributes:
        data (torch.Tensor): 数据
        target (torch.Tensor): 目标值
    """

    def __init__(self, signals, Kalman_H):
        """构造函数
        
        Args:
            signals (np.array): 原始信号数据
            Kalman_H (np.array): Kalman 候选项矩阵
        """

        super().__init__()
        n_dim, n_point, _ = Kalman_H.shape
        max_lag = signals.shape[0] - Kalman_H.shape[1]
        data = []
        for t in range(n_point):
            Cn = np.kron(np.eye(n_dim), Kalman_H[:, t, :])
            data.append(Cn[slice(0, Cn.shape[0], n_dim + 1)])
        self.data = torch.as_tensor(np.array(data)).float()
        self.target = torch.as_tensor(signals[max_lag:]).float()

    def __getitem__(self, index):
        return self.data[index], self.target[index]

    def __len__(self):
        return self.data.size(0)


def make_dataset4SK(signals, Kalman_H):
    """构造用于 scikit-learn 使用的数据集
        
    Args:
        signals (np.array): 原始信号数据
        Kalman_H (np.array): Kalman 候选项矩阵
    """

    n_dim, n_point, _ = Kalman_H.shape
    max_lag = signals.shape[0] - Kalman_H.shape[1]
    data = []
    for t in range(n_point):
        Cn = np.kron(np.eye(n_dim), Kalman_H[:, t, :])
        data.append(Cn[slice(0, Cn.shape[0], n_dim + 1)])
    return np.array(data), signals[max_lag:]
